from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from .models.customers import Customer
from .models.meta import Empty
from .models.response import (
    BillingPortalResponse,
    ListCustomerResponse,
    PricingTableResponse,
)
from .utils import _build_payload

if TYPE_CHECKING:
    from .aio.http import AsyncHTTPClient
    from .http import HTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")

__all__ = ("Customers",)


class Customers(Generic[T_HttpClient]):
    """An interface to Autumn's customer API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.
    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def get(
        self: "Customers[HTTPClient]",
        customer_id: str,
        expand: Optional[
            List[
                Literal[
                    "invoices",
                    "rewards",
                    "trials_used",
                    "entities",
                    "referrals",
                    "payment_method",
                ]
            ]
        ] = None,
    ) -> Customer: ...

    @overload
    def get(
        self: "Customers[AsyncHTTPClient]",
        customer_id: str,
        expand: Optional[
            List[
                Literal[
                    "invoices",
                    "rewards",
                    "trials_used",
                    "entities",
                    "referrals",
                    "payment_method",
                ]
            ]
        ] = None,
    ) -> Coroutine[Any, Any, Customer]: ...

    def get(
        self,
        customer_id: str,
        expand: Optional[
            List[
                Literal[
                    "invoices",
                    "rewards",
                    "trials_used",
                    "entities",
                    "referrals",
                    "payment_method",
                ]
            ]
        ] = None,
    ) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Get a customer by their ID.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get.
        expand: Optional[List[Literal["invoices", "rewards", "trials_used", "entities", "referrals", "payment_method"]]]
            Additional fields to expand in the response.

        Returns
        -------
        :class:`~autumn.models.customers.Customer`
            The customer.
        """
        params = {}
        if expand is not None:
            params["expand"] = expand
        return self._http.request(
            "GET", f"/customers/{customer_id}", Customer, params=params
        )

    @overload
    def list(
        self: "Customers[HTTPClient]", *, limit: int = 10, offset: int = 0
    ) -> ListCustomerResponse: ...

    @overload
    def list(
        self: "Customers[AsyncHTTPClient]", *, limit: int = 10, offset: int = 0
    ) -> Coroutine[Any, Any, ListCustomerResponse]: ...

    def list(
        self, *, limit: int = 10, offset: int = 0
    ) -> Union[
        ListCustomerResponse, Coroutine[Any, Any, ListCustomerResponse]
    ]:
        params = {"limit": limit, "offset": offset}
        return self._http.request(
            "GET", "/customers", ListCustomerResponse, params=params
        )

    @overload
    def create(
        self: "Customers[HTTPClient]",
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        stripe_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Customer: ...

    @overload
    def create(
        self: "Customers[AsyncHTTPClient]",
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        stripe_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Coroutine[Any, Any, Customer]: ...

    def create(
        self,
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        stripe_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Create a new customer.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the customer to create.
        email: Optional[str]
            The customer's email address.
        name: Optional[str]
            The customer's name.
        stripe_id: Optional[str]
            The customer's Stripe ID. This can be a new or existing ID.
        metadata: Optional[Dict[str, Any]]
            Additional metadata to attach to the customer.

        Returns
        -------
        :class:`~autumn.models.customers.Customer`
            The created customer.
        """
        payload = _build_payload(locals(), Customers.create)
        return self._http.request("POST", "/customers", Customer, json=payload)

    @overload
    def update(
        self: "Customers[HTTPClient]",
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Customer: ...

    @overload
    def update(
        self: "Customers[AsyncHTTPClient]",
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Coroutine[Any, Any, Customer]: ...

    def update(
        self,
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Update a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to update.
        name: Optional[str]
            The customer's new name.
        email: Optional[str]
            The customer's new email address.
        fingerprint: Optional[str]
            The customer's new fingerprint.

        Returns
        -------
        :class:`~autumn.models.customers.Customer`
            The updated customer.
        """
        payload = _build_payload(
            locals(), Customers.update, ignore={"customer_id"}
        )
        return self._http.request(
            "POST", f"/customers/{customer_id}", Customer, json=payload
        )

    @overload
    def delete(self: "Customers[HTTPClient]", customer_id: str) -> Empty: ...

    @overload
    def delete(
        self: "Customers[AsyncHTTPClient]", customer_id: str
    ) -> Coroutine[Any, Any, Empty]: ...

    def delete(
        self, customer_id: str
    ) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Delete a customer.

        Parameters
        ----------
        customer_id: str
            The ID of the customer to delete.

        Returns
        -------
        :class:`~autumn.models.customers.Empty`
            An empty response.

        |maybecoro|
        """
        return self._http.request("DELETE", f"/customers/{customer_id}", Empty)

    @overload
    def get_billing_portal(
        self: "Customers[HTTPClient]",
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> BillingPortalResponse: ...

    @overload
    def get_billing_portal(
        self: "Customers[AsyncHTTPClient]",
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> Coroutine[Any, Any, BillingPortalResponse]: ...

    def get_billing_portal(
        self,
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> Union[
        BillingPortalResponse, Coroutine[Any, Any, BillingPortalResponse]
    ]:
        """Get a billing portal URL for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a billing portal URL for.
        return_url: Optional[str]
            The URL to return to after the customer has completed the billing portal.

        Returns
        -------
        :class:`~autumn.models.response.BillingPortalResponse`
            The billing portal URL.
        """
        payload = _build_payload(
            locals(),
            Customers.get_billing_portal,
            ignore={"customer_id"},
        )
        return self._http.request(
            "POST",
            f"/customers/{customer_id}/billing_portal",
            BillingPortalResponse,
            json=payload,
        )

    @overload
    def pricing_table(
        self: "Customers[HTTPClient]", customer_id: str
    ) -> PricingTableResponse: ...

    @overload
    def pricing_table(
        self: "Customers[AsyncHTTPClient]", customer_id: str
    ) -> Coroutine[Any, Any, PricingTableResponse]: ...

    def pricing_table(
        self, customer_id: str
    ) -> Union[
        PricingTableResponse, Coroutine[Any, Any, PricingTableResponse]
    ]:
        """Get a pricing table for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a pricing table for.

        Returns
        -------
        :class:`~autumn.models.response.PricingTableResponse`
            The pricing table.
        """
        params = {"customer_id": customer_id}
        return self._http.request(
            "GET",
            "/components/pricing_table",
            PricingTableResponse,
            params=params,
        )



import sys
import time
from typing import Dict, Type, TypeVar

import requests
from pydantic import BaseModel

from .error import AutumnError, AutumnHTTPError
from .utils import ExponentialBackoff, _build_model, _check_response

__all__ = ("HTTPClient",)

T = TypeVar("T", bound=BaseModel)


class _RetryRequestError(Exception):
    pass


class HTTPClient:
    def __init__(self, base_url: str, version: str, token: str, attempts: int):
        self.base_url = base_url
        self.version = version
        self.session = requests.Session()

        self._headers = self._build_headers(token)
        self.attempts = attempts

    @staticmethod
    def _build_url(base_url: str, version: str, path: str) -> str:
        return f"{base_url}/{version}{path}"

    @staticmethod
    def _build_headers(token: str) -> Dict[str, str]:
        from . import LATEST_API_VERSION, __version__

        v_info = sys.version_info
        user_agent = (
            f"autumn-py/{__version__}"
            f" (Python {v_info.major}.{v_info.minor}.{v_info.micro})"
        )
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": user_agent,
            "Authorization": f"Bearer {token}",
            "X-Api-Version": LATEST_API_VERSION,
        }
        return headers

    def request(
        self,
        method: str,
        path: str,
        type_: Type[T],
        **kwargs,
    ) -> T:
        if self.session is None:
            raise AutumnError(
                "Session is not initialized. You may have closed the client accidentally.",
                "session_not_initialized",
            )

        url = self._build_url(self.base_url, self.version, path)

        max_attempts = self.attempts
        backoff = ExponentialBackoff()
        for attempt in range(max_attempts):
            try:
                resp = self.session.request(
                    method, url, headers=self._headers, **kwargs
                )
                if 500 <= resp.status_code <= 504:
                    raise _RetryRequestError()

                data = resp.json()
            except (
                _RetryRequestError,
                OSError,
                requests.ConnectionError,
                requests.ConnectTimeout,
            ):
                if attempt == max_attempts - 1:
                    raise

                time.sleep(backoff.bedtime)
                backoff.tick()
            else:
                _check_response(resp.status_code, data)
                return _build_model(type_, data)

        # We should never get here. This is to appease type checkers.
        msg = f"Max retries reached for {method} {path}"
        raise AutumnHTTPError(msg, "max_retries_reached", 500)

    def close(self):
        if self.session is not None:
            self.session.close()
            self.session = None



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from autumn.models.response import CheckoutResponse

if TYPE_CHECKING:
    from typing import Any, Awaitable, Dict, List, Literal, Optional, Union

    from ..models.meta import CustomerData, FeatureOptions, ProductOptions
    from ..models.response import (
        AttachResponse,
        CancelResponse,
        CheckoutResponse,
        CheckResponse,
        QueryResponse,
        TrackResponse,
    )

__all__ = ("AttachParams", "CheckParams", "TrackParams", "CancelParams")


class AttachParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        product_ids: Optional[List[str]] = None,
        products: Optional[List[ProductOptions]] = None,
        success_url: Optional[str] = None,
        force_checkout: bool = False,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
        free_trial: Optional[bool] = None,
        options: Optional[List[FeatureOptions]] = None,
        reward: Optional[Union[str, List[str]]] = None,
    ) -> Awaitable[AttachResponse]: ...


class CheckParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        required_balance: Optional[int] = 1,
        send_event: bool = False,
        with_preview: bool = False,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> Awaitable[CheckResponse]: ...


class TrackParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        feature_id: Optional[str] = None,
        value: int = 1,
        entity_id: Optional[str] = None,
        event_name: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> Awaitable[TrackResponse]: ...


class CheckoutParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        products: Optional[List[ProductOptions]] = None,
        entity_id: Optional[str] = None,
        options: Optional[List[FeatureOptions]] = None,
        customer_data: Optional[CustomerData] = None,
        success_url: Optional[str] = None,
        checkout_session_params: Optional[Dict[str, Any]] = None,
        reward: Optional[Union[str, List[str]]] = None,
    ) -> Awaitable[CheckoutResponse]: ...


class QueryParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        feature_id: Union[str, List[str]],
        *,
        range: Literal["24h", "7d", "30d", "90d", "last_cycle"] = "30d",
    ) -> Awaitable[QueryResponse]: ...


class CancelParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        product_id: str,
        *,
        entity_id: Optional[str] = None,
        cancel_immediately: bool = False,
    ) -> Awaitable[CancelResponse]: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Coroutine, TypedDict

try:
    from starlette.middleware import Middleware
    from starlette.middleware.base import BaseHTTPMiddleware
    from starlette.responses import JSONResponse
    from starlette.routing import Route, Router
except ImportError:
    from ..error import AutumnError

    STARLETTE_INSTALLED = False

    raise AutumnError(
        "starlette is not installed. Please install it with `pip install starlette`",
        "missing_dependency",
    )
else:
    STARLETTE_INSTALLED = True

from ..aio.client import AsyncClient
from ..error import AutumnHTTPError
from .routes.core import (
    attach_route,
    billing_portal_route,
    cancel_route,
    check_route,
    checkout_route,
    query_route,
    track_route,
)
from .routes.customers import create_customer_route, pricing_table_route
from .routes.entities import (
    create_entity_route,
    delete_entity_route,
    get_entity_route,
)
from .routes.products import list_products_route

if TYPE_CHECKING:
    from starlette.requests import Request


class _CustomerData(TypedDict):
    name: str
    email: str


class AutumnIdentifyData(TypedDict):
    customer_id: str
    customer_data: _CustomerData


class _StateMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, *, client, identify):
        super().__init__(app)
        self._client = client
        self._identify = identify

    async def dispatch(self, request: Request, call_next):
        request.state.__autumn__ = {
            "client": self._client,
            "identify": self._identify,
        }
        return await call_next(request)


class AutumnASGI:
    def __init__(
        self,
        token: str,
        *,
        identify: Callable[[Request], Coroutine[Any, Any, AutumnIdentifyData]],
    ):
        self._client = AsyncClient(token)
        self._identify = identify

        router = Router(
            routes=[
                Route("/attach/", attach_route, methods={"POST"}),
                Route("/check/", check_route, methods={"POST"}),
                Route("/track/", track_route, methods={"POST"}),
                Route("/cancel/", cancel_route, methods={"POST"}),
                Route("/query/", query_route, methods={"POST"}),
                Route(
                    "/products/",
                    list_products_route,
                    methods={"GET"},
                ),
                Route(
                    "/billing_portal/",
                    billing_portal_route,
                    methods={"POST"},
                ),
                Route("/checkout/", checkout_route, methods={"POST"}),
                Route("/customers/", create_customer_route, methods={"POST"}),
                Route(
                    "/customers/{customer_id:str}/entities/{entity_id:str}",
                    delete_entity_route,
                    methods={"DELETE"},
                ),
                Route(
                    "/customers/{customer_id:str}/entities/",
                    create_entity_route,
                    methods={"POST"},
                ),
                Route(
                    "/customers/{customer_id:str}/entities/{entity_id:str}/",
                    get_entity_route,
                    methods={"GET"},
                ),
                Route(
                    "/components/pricing_table/",
                    pricing_table_route,
                    methods={"GET"},
                ),
            ],
            middleware=[
                Middleware(
                    _StateMiddleware,
                    client=self._client,
                    identify=self._identify,
                )
            ],
        )
        self._router = router

    async def _handle_http_error(self, _: Request, exc: AutumnHTTPError):
        return JSONResponse({"message": f"{exc.message} ({exc.code})"})

    async def close(self):
        await self._client.close()

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        try:
            await self._router(scope, receive, send)
        except AutumnHTTPError as exc:
            response = JSONResponse(
                {"detail": str(exc)}, status_code=exc.status_code or 400
            )
            await response(scope, receive, send)



from typing import List, Optional

from pydantic import BaseModel

from .customers import Customer, PricingTableProduct
from .env import AppEnv
from .features import FeaturePreview
from .meta import FeatureOptions, QueryDP
from .products import FreeTrial, Product, ProductItem, ProductPreview

__all__ = ("AttachResponse", "CheckResponse")


class AttachResponse(BaseModel):
    customer_id: str
    code: str = "No code returned."
    message: str
    checkout_url: Optional[str] = None
    product_ids: List[str]
    success: Optional[bool] = None


class CheckResponse(BaseModel):
    allowed: bool
    customer_id: str
    code: str
    balance: Optional[float] = None
    feature_id: Optional[str] = None
    product_id: Optional[str] = None
    status: Optional[str] = None
    feature_preview: Optional[FeaturePreview] = None
    product_preview: Optional[ProductPreview] = None


class TrackResponse(BaseModel):
    id: str
    code: str
    customer_id: str
    feature_id: Optional[str] = None
    event_name: Optional[str] = None


class UsageResponse(BaseModel):
    code: str
    customer_id: str
    feature_id: str


class BillingPortalResponse(BaseModel):
    url: str
    customer_id: str


class CreateProductResponse(BaseModel):
    created_at: int
    id: str
    name: str
    env: AppEnv
    version: int
    is_add_on: bool
    is_default: bool
    items: List[ProductItem]
    free_trial: Optional[FreeTrial] = None


GetProductResponse = CreateProductResponse


class ReferralCodeResponse(BaseModel):
    code: str
    customer_id: str
    created_at: int


class ReferralRedeemResponse(BaseModel):
    id: str
    customer_id: str
    reward_id: str


class CancelResponse(BaseModel):
    success: bool
    customer_id: str
    product_id: str


class PricingTableResponse(BaseModel):
    list: List[PricingTableProduct]


class CheckoutLine(BaseModel):
    description: str
    amount: float
    item: ProductItem


class CheckoutNextCycle(BaseModel):
    starts_at: int
    total: float


class CheckoutResponse(BaseModel):
    url: Optional[str] = None
    customer_id: str
    has_prorations: bool
    lines: List[CheckoutLine]
    total: float
    currency: str
    options: List[FeatureOptions]
    product: Optional[Product] = None
    current_product: Optional[Product] = None
    next_cycle: Optional[CheckoutNextCycle] = None


class ListProductResponse(BaseModel):
    list: List[Product]


class ListCustomerResponse(BaseModel):
    list: List[Customer]
    total: int
    limit: int
    offset: int


class QueryResponse(BaseModel):
    list: List[QueryDP]


class TransferProductResponse(BaseModel):
    success: bool



import pytest

from autumn.models.customers import CustomerFeature, ProductItemInterval
from autumn.customers import Customers
from autumn.error import AutumnValidationError, AutumnHTTPError
from autumn.utils import _build_model, _build_payload, _decompose_value, _check_response


def test_decompose_value_singular():
    feature = CustomerFeature(
        id="mock_id",
        name="John Doe",
        unlimited=False,
        interval=ProductItemInterval.HOUR,
    )
    decomposed_feature = {
        "id": "mock_id",
        "name": "John Doe",
        "unlimited": False,
        "interval": "hour",
        "balance": None,
        "usage": None,
        "included_usage": None,
        "next_reset_at": None,
        "breakdown": None,
        "type": None,
    }

    assert _decompose_value(feature) == decomposed_feature


def test_decompose_value_recursive():
    feature = CustomerFeature(
        id="mock_id",
        name="John Doe",
        unlimited=False,
        interval=ProductItemInterval.HOUR,
    )
    decomposed_feature = {
        "id": "mock_id",
        "name": "John Doe",
        "unlimited": False,
        "interval": "hour",
        "balance": None,
        "usage": None,
        "included_usage": None,
        "next_reset_at": None,
        "breakdown": None,
        "type": None,
    }

    features = [feature] * 50
    decomposed_features = [decomposed_feature] * 50

    assert _decompose_value(features) == decomposed_features


def test_build_payload():
    mock_scope = {
        "self": None,
        "id": "mock_customer_id",
        "email": "johndoe@example.com",
        "name": "John Doe",
        **locals(),
    }
    expected_payload = {
        "id": "mock_customer_id",
        "email": "johndoe@example.com",
        "name": "John Doe",
    }

    assert _build_payload(mock_scope, Customers.create) == expected_payload


def test_model_build_exc():
    bad_data = {
        "id": "mock_id",
        "name": "John Doe",
        "unlimited": False,
        "interval": "chocolate",
        "balance": "",
        "usage": None,
        "included_usage": None,
        "next_reset_at": None,
        "breakdown": None,
    }

    with pytest.raises(AutumnValidationError):
        _build_model(CustomerFeature, bad_data)


def test_check_response():
    data = {"mock": "data"}

    for code in range(200, 299 + 1):
        _check_response(code, data)

    with pytest.raises(AutumnHTTPError):
        _check_response(300, data)

    with pytest.raises(AutumnHTTPError):
        _check_response(301, data)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from ..client import Client
from ..customers import Customers
from ..entities import Entities
from ..error import AutumnError
from ..features import Features
from ..products import Products
from .http import AsyncHTTPClient

try:
    import aiohttp
except ImportError:
    raise AutumnError(
        "aiohttp is not installed. Please install it with `pip install aiohttp`",
        "missing_dependency",
    )

if TYPE_CHECKING:
    from typing_extensions import Self

    from .stubs import (
        AttachParams,
        CancelParams,
        CheckoutParams,
        CheckParams,
        QueryParams,
        TrackParams,
    )

__all__ = ("AsyncClient",)


class AsyncClient(Client):
    """
    The ``async`` client class for interacting with the Autumn API.

    This class is also exposed as ``autumn.Autumn``.

    The ``AsyncClient`` automatically retries requests up to 5 times, exponentially backing off between attempts.

    Note that session creation is lazy. This means that the ``AsyncClient`` will not attempt to create a session until the first request is made.

    Parameters
    ----------
    token: str
        The API key to use for authentication.
    base_url: Optional[str]
        The base URL of the Autumn API. This is useful when you are self-hosting Autumn and need to point to your own instance.
    max_retries: int
        The maximum number of retries to attempt for failed requests.
    session: Optional[:class:`~aiohttp.ClientSession`]
        The session to use for requests. If not provided, a new session will be created **lazily**.

    Attributes
    ----------
    customers: :class:`~autumn.customers.Customers`
        An interface to Autumn's customer API.
    features: :class:`~autumn.features.Features`
        An interface to Autumn's feature API.
    products: :class:`~autumn.products.Products`
        An interface to Autumn's product API.
    entities: :class:`~autumn.entities.Entities`
        An interface to Autumn's entities API.

    """

    attach: AttachParams  # type: ignore
    check: CheckParams  # type: ignore
    track: TrackParams  # type: ignore
    checkout: CheckoutParams  # type: ignore
    query: QueryParams  # type: ignore
    cancel: CancelParams  # type: ignore

    def __init__(
        self,
        token: str,
        *,
        base_url: Optional[str] = None,
        max_retries: int = 5,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        from .. import BASE_URL, VERSION

        _base_url = base_url or BASE_URL
        _base_url = _base_url.rstrip("/")

        attempts = max_retries + 1
        self.http = AsyncHTTPClient(
            _base_url,
            VERSION,
            token,
            attempts=attempts,
            session=session,
        )
        self.customers = Customers(self.http)
        self.features = Features(self.http)
        self.products = Products(self.http)
        self.entities = Entities(self.http)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, _exc_type, _exc, _tb):
        await self.close()

    async def close(self):
        await self.http.close()



from typing import Dict, List, Literal, Optional

from pydantic import BaseModel

from .customers import CustomerFeature, CustomerInvoice, CustomerProduct
from .products import Product

__all__ = ("Feature", "FeaturePreview", "Entity")


class Feature(BaseModel):
    feature_id: str
    quantity: int


class FeaturePreview(BaseModel):
    title: str
    message: str
    scenario: Literal["usage_limit", "feature_flag"]
    feature_id: str
    feature_name: str
    upgrade_product_id: str
    products: List[Product]


class Entity(BaseModel):
    id: str
    name: str
    customer_id: str
    created_at: int
    env: str
    products: List[CustomerProduct]
    features: Dict[str, CustomerFeature]
    invoices: Optional[List[CustomerInvoice]] = None



from typing import Any, List

from typing_extensions import Self

__all__ = ("AutumnError", "AutumnValidationError")


class AutumnError(Exception):
    """
    Base exception for all Autumn errors.

    Attributes
    ----------
    message: str
        The error message.
    code: str
        The error code.
    """

    def __init__(self, message: str, code: str):
        self.message = message
        self.code = code

        if not hasattr(self, "__notes__"):
            self.__notes__: List[str] = []

    def __str__(self):
        return f"{self.code}: {self.message}"

    def __repr__(self):
        notes = "\n".join(self.__notes__)
        return f"{self.__class__.__name__}({self.message}, {self.code}){notes}"

    def add_note(self, note: str):
        self.__notes__.append(note)


class AutumnValidationError(AutumnError):
    """
    Exception raised when a validation error occurs.
    This is raised when the API returns a response that isn't recognized by the library.

    Attributes
    ----------
    message: str
        The error message.
    code: str
        The error code.
    """

    def __init__(self, message: str, code: str):
        super().__init__(message, code)


class AutumnHTTPError(AutumnError):
    """
    Exception raised when an HTTP error occurs.

    Attributes
    ----------
    message: str
        The error message. This can be an empty string.
    code: str
        The error code.
    status_code: int
        The HTTP status code.
    """

    def __init__(self, message: str, code: str, status_code: int):
        super().__init__(message, code)
        self.status_code = status_code

        self.body = ""

    def __str__(self):
        return f"{self.code!r}: {self.message!r} (HTTP {self.status_code})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message}, {self.code}, {self.status_code})\n{self.body}"

    def attach_body(self, body: Any) -> Self:
        self.body = body
        return self



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)

from .models.meta import Empty
from .models.products import FreeTrial, ProductItem
from .models.response import (
    CreateProductResponse,
    GetProductResponse,
    ListProductResponse,
    ReferralCodeResponse,
    ReferralRedeemResponse,
)
from .utils import _build_payload

if TYPE_CHECKING:
    from .aio.http import AsyncHTTPClient
    from .http import HTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")


class Products(Generic[T_HttpClient]):
    """An interface to Autumn's product API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.

    Example:

    .. code-block:: python

        import autumn

        client = autumn.Client(token="your_api_key")

        product = client.products.create(
            id="chat_messages",
            name="Chat Messages",
        )

        print(product.name)  # Chat Messages

    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def create(
        self: "Products[HTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> CreateProductResponse: ...

    @overload
    def create(
        self: "Products[AsyncHTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Coroutine[Any, Any, CreateProductResponse]: ...

    def create(
        self,
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Union[
        CreateProductResponse, Coroutine[Any, Any, CreateProductResponse]
    ]:
        """Create a new product.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product.
        name: str
            The name of the product.
        is_add_on: bool
            Whether the product is an add-on.
        is_default: bool
            Whether the product is a default product.
        items: Optional[List[ProductItem]]
            The items of the product.
        free_trial: Optional[FreeTrial]
            The free trial configuration of the product.

        Returns
        -------
        :class:`~autumn.models.response.CreateProductResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.create)  # type: ignore
        return self._http.request(
            "POST", "/products", CreateProductResponse, json=payload
        )

    @overload
    def get(
        self: "Products[HTTPClient]",
        id: str,
    ) -> GetProductResponse: ...

    @overload
    def get(
        self: "Products[AsyncHTTPClient]",
        id: str,
    ) -> Coroutine[Any, Any, GetProductResponse]: ...

    def get(
        self, id: str
    ) -> Union[GetProductResponse, Coroutine[Any, Any, GetProductResponse]]:
        """Get a product by its ID.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product to get.

        Returns
        -------
        :class:`~autumn.models.response.GetProductResponse`
            The response from the API.
        """
        return self._http.request("GET", f"/products/{id}", GetProductResponse)

    @overload
    def list(
        self: "Products[HTTPClient]", customer_id: str
    ) -> ListProductResponse: ...

    @overload
    def list(
        self: "Products[AsyncHTTPClient]", customer_id: str
    ) -> Coroutine[Any, Any, ListProductResponse]: ...

    def list(
        self, customer_id: str
    ) -> Union[ListProductResponse, Coroutine[Any, Any, ListProductResponse]]:
        """List products for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to list products for.

        Returns
        -------
        :class:`~autumn.models.response.ListProductResponse`
            The response from the API.
        """
        params = {"customer_id": customer_id}
        return self._http.request(
            "GET", "/products", ListProductResponse, params=params
        )

    @overload
    def update(
        self: "Products[HTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Empty: ...

    @overload
    def update(
        self: "Products[AsyncHTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Coroutine[Any, Any, Empty]: ...

    def update(
        self,
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Update a product.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product to update.
        name: str
            The new name of the product.
        is_add_on: bool
            Whether the product is an add-on.
        is_default: bool
            Whether the product is the default product.
        items: Optional[List[ProductItem]]
            The new items of the product.
        free_trial: Optional[FreeTrial]
            The new free trial of the product.

        Returns
        -------
        :class:`~autumn.models.response.Empty`
            This is a placeholder type. Treat it as :class:`None`.
        """

        payload = _build_payload(locals(), self.update)  #  type: ignore
        return self._http.request(
            "POST", f"/products/{id}", Empty, json=payload
        )

    @overload
    def delete(self: "Products[HTTPClient]", id: str) -> Empty: ...

    @overload
    def delete(self: "Products[AsyncHTTPClient]", id: str) -> Empty: ...

    def delete(self, id: str) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Delete a product.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product to delete.

        Returns
        -------
        :class:`~autumn.models.response.Empty`
            This is a placeholder type. Treat it as :class:`None`.
        """
        return self._http.request("DELETE", f"/products/{id}", Empty)

    @overload
    def get_referral_code(
        self: "Products[HTTPClient]",
        customer_id: str,
        program_id: str,
    ) -> ReferralCodeResponse: ...

    @overload
    def get_referral_code(
        self: "Products[AsyncHTTPClient]",
        customer_id: str,
        program_id: str,
    ) -> Coroutine[Any, Any, ReferralCodeResponse]: ...

    def get_referral_code(
        self,
        customer_id: str,
        program_id: str,
    ) -> Union[
        ReferralCodeResponse, Coroutine[Any, Any, ReferralCodeResponse]
    ]:
        """Get a referral code for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a referral code for.
        program_id: str
            The ID of the program to get a referral code for.

        Returns
        -------
        :class:`~autumn.models.response.ReferralCodeResponse`
            The response from the API.
        """

        payload = _build_payload(locals(), self.get_referral_code)  # type: ignore
        return self._http.request(
            "POST",
            "/referrals/code",
            ReferralCodeResponse,
            json=payload,
        )

    @overload
    def redeem_referral_code(
        self: "Products[HTTPClient]",
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> ReferralRedeemResponse: ...

    @overload
    def redeem_referral_code(
        self: "Products[AsyncHTTPClient]",
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> Coroutine[Any, Any, ReferralRedeemResponse]: ...

    def redeem_referral_code(
        self,
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> Union[
        ReferralRedeemResponse, Coroutine[Any, Any, ReferralRedeemResponse]
    ]:
        """Redeem a referral code for a customer.

        |maybecoro|

        Parameters
        ----------
        code: str
            The code to redeem.
        customer_id: str
            The ID of the customer to redeem the code for.
        reward_id: str
            The ID of the reward to redeem.

        Returns
        -------
        :class:`~autumn.models.response.ReferralRedeemResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.redeem_referral_code)  # type: ignore
        return self._http.request(
            "POST",
            "/referrals/redeem",
            ReferralRedeemResponse,
            json=payload,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from . import _build_response, _extract

if TYPE_CHECKING:
    from starlette.requests import Request


async def create_entity_route(request: Request):
    _, autumn, json = await _extract(request)

    customer_id = request.path_params["customer_id"]
    response = await autumn.entities.create(customer_id=customer_id, **json)

    return _build_response(response)


async def delete_entity_route(request: Request):
    _, autumn, _ = await _extract(request)

    customer_id = request.path_params["customer_id"]
    entity_id = request.path_params["entity_id"]
    response = await autumn.entities.delete(customer_id, entity_id)

    return _build_response(response)


async def get_entity_route(request: Request):
    _, autumn, _ = await _extract(request)

    customer_id = request.path_params["customer_id"]
    entity_id = request.path_params["entity_id"]
    expand = request.query_params.get("expand", None)

    if expand is not None:
        expand = expand.split(",")

    response = await autumn.entities.get(customer_id, entity_id, expand=expand)  # type: ignore
    return _build_response(response)



from .aio.client import AsyncClient as Autumn
from .client import *
from .error import *
from .models.balance import *
from .models.customers import *
from .models.entities import *
from .models.features import *
from .models.meta import *
from .models.products import *

__title__ = "autumn"
__version__ = "3.1.1"
__license__ = "MIT"
__author__ = "justanotherbyte"
__copyright__ = "Copyright 2025 justanotherbyte"

BASE_URL = "https://api.useautumn.com"
VERSION = "v1"
LATEST_API_VERSION = "1.4"



from __future__ import annotations

from typing import TYPE_CHECKING

from ...models.meta import CustomerData
from ...utils import _build_payload as _build_kwargs
from . import _build_response, _extract

if TYPE_CHECKING:
    from starlette.requests import Request


async def attach_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    customer_data = identify["customer_data"]

    kwargs = _build_kwargs(json, autumn.attach)

    response = await autumn.attach(
        customer_id, customer_data=CustomerData(**customer_data), **kwargs
    )
    return _build_response(response)


async def check_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(json, autumn.check)

    response = await autumn.check(customer_id, **kwargs)
    return _build_response(response)


async def track_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(json, autumn.track)

    response = await autumn.track(customer_id, **kwargs)
    return _build_response(response)


async def cancel_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(json, autumn.cancel)

    response = await autumn.cancel(customer_id, **kwargs)
    return _build_response(response)


async def billing_portal_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(
        json,
        autumn.customers.get_billing_portal,
    )
    response = await autumn.customers.get_billing_portal(customer_id, **kwargs)
    return _build_response(response)


async def checkout_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(json, autumn.checkout)

    response = await autumn.checkout(customer_id, **kwargs)
    return _build_response(response)


async def query_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    kwargs = _build_kwargs(json, autumn.query)

    response = await autumn.query(customer_id=customer_id, **kwargs)
    return _build_response(response)
